/*	$Id: matrix.c 2007/08/27 12:10:13 mallet $ */

/*
 * Copyright (c) 2007 Anthony Mallet                  --  Thu Aug 23 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <err.h>

#include "libpom/lib.h"
#include "pomInt.h"

#define TOLERANCE	1e-4
#define ITERATIONS	10000
#define SRAND		time(NULL)

#define MATELT(mat, i, j, c)	(*((mat) + (c) * (i) + (j)) )

#define INDEX_SMATELT(i, j)   \
   ((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j))
#define SMATELT(mat, i, j, n)	(*((mat) + INDEX_SMATELT(i,j)))

int
main(int argc, const char *argv[])
{
  double a[21], s[21], l[21];
  int iter, k, i, j, r;
  double val;

  srand48(SRAND);
  for(iter = 0; iter < ITERATIONS; iter++) {

    /* make a a random symmetric matrix */
    for(i=0; i<21; i++) a[i] = 10*drand48();

    /* make s positive definite by computing s = a x a */
    for (i=0; i<6; i++) for (j=i; j<6; j++) {
      val = 0.;
      for (k=0; k<6; k++)
	val += SMATELT(a, i, k, 6) * SMATELT(a, k, j, 6);
      SMATELT(s, i, j, 6) = val;
    }

    /* invert s in l */
    r = pom_dms_inv_6x6(s, l);
    if (r) { warnx("pom_dms_inv_6x6 failed"); exit(2); }

    /* compute product a = s x l */
    for (i=0; i<6; i++) for (j=i; j<6; j++) {
      val = 0.;
      for (k=0; k<6; k++)
	val += SMATELT(s, i, k, 6) * SMATELT(l, k, j, 6);
      SMATELT(a, i, j, 6) = val;
    }

    /* a == identity */
    for (i=0; i<6; i++) for (j=i; j<6; j++)
      if (fabs(SMATELT(a, i, j, 6) - (i==j)) > TOLERANCE) {
	for (i=0; i<6; i++) {
	  for (j=i; j<6; j++) printf("%g ", SMATELT(a, i, j, 6));
	  printf("\n");
	}
	warnx("pom_dms_inv_6x6 failed, iteration %d", iter);
	exit(2);
      }
  }

  warnx("pom_dms_inv_6x6 succeeded on %d random matrices.", iter);
  return 0;
}
