/*	$Id: fusion.c 2007/08/24 19:40:12 mallet $ */

/*
 * Copyright (c) 2003-2007 LAAS/CNRS                  --  Fri Sep  1 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: fusion.c 2007/08/24 19:40:12 mallet $";

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "t3d/t3d.h"

#include "libpom/lib.h"
#include "pomInt.h"



/**
 * pomUpdateB2O ---------------------------------------------------------
 *
 * Update the BaseToOrigin matrix of every ME whom confidence is less that
 * the confidence of the ME doing the update
 * To sum it up, BaseToOrigin must hold the transform to the best current pose
 * for normal poses, to the Ref one for me[0]
 */

int
pomUpdateB2O(PomMESet *set, int me, int index,T3D *m2o)
{
  int i;
   double conf = set->me[me].p[index]->pos.confidence;

   t3dInit(m2o, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);

   /** Compute pos */
   if (set->me[me].p[index]->pos.class == POM_ABSOLUTE_ME) {
     /** If the updater ME is absolute, get its pose and covariance */
     t3dCopyIn(m2o, &set->me[me].p[index]->pos.pos.abs.mainToOrigin);
   }
   else {

     /** Else get the corresponding reference pose
        and computes the covariance matrix in consequence
        Removing the correlation terms
        See note from Ph. Moutarlier : it can be generalized
        as in the following lines */
#ifdef POM_DEBUG
     fprintf(stderr,"Using odo (%d) for index %d \n",
             me,index);
#endif
     /* init */
     if (set->me[me].lastabs.date==100){
       t3dCopyIn(m2o, &set->me[me].p[index]->pos.pos.abs.mainToOrigin);
     } else {
       if (pomIntegrateBaseToOrigin(m2o,set,me,index)) {
	 printf("pomIntegrateBaseToOrigin failed \n");
	 if (set->log!=NULL) gzprintf(set->log,
				      "pomIntegrateBaseToOrigin failed \n");
	 return 1;
       }
     }
   }

   /** update other's b2o */
   for(i=1;i<set->nmes;i++) {
     if (i != me && set->me[i].fuse &&
	 set->me[i].p[index] != NULL &&
	 set->me[i].p[index]->pos.confidence < conf) {
#ifdef POM_DEBUG
      fprintf(stderr,"Update from %d to %d\n",me,i);
#endif
      if (set->log!=NULL) gzprintf(set->log,
				    "Update from %d to %d\n",me,i);
        /** Copy MainToOrigin */
      set->me[i].lastabs.date = set->me[i].p[index]->pos.pos.abs.date;

      t3dCopyIn(&set->me[i].lastabs.mainToOrigin,&set->me[i].p[index]->pos.pos.abs.mainToOrigin);

      t3dCopyIn(&set->me[i].baseToOrigin,&set->me[i].p[index]->pos.pos.abs.mainToOrigin);

      /** Inverts it */
      t3dInvert(&set->me[i].baseToOrigin);
      /** Compose it with with the updater ME pose
	  The covariance matrix is also updated in the process */
      t3dComp(&set->me[i].baseToOrigin, m2o);

      t3dConvertTo(T3D_U_MATRIX,&set->me[i].baseToOrigin);
      t3dSetOk(&set->me[i].baseToOrigin);
      t3dConvertTo(T3D_U_MATRIX,m2o);
      set->me[me].lastabs.date = set->me[me].p[index]->pos.pos.abs.date;



      t3dConvertTo(T3D_U_MATRIX,&set->me[me].baseToOrigin);

      /** MAJ de la date de derniere loc */
      }
   }
   return 0;
}


/**
 * pomKalmanUpdateB2O ---------------------------------------------------------
 *
 * Update the BaseToOrigin matrix of every ME whom confidence is less that
 * the confidence of the ME doing the update
 * To sum it up, BaseToOrigin must hold the transform to the best current pose
 * for normal poses, to the Ref one for me[0]
 */

#define NB_MAX_ME 10

int
pomKalmanUpdateB2O(PomMESet *set, int index, int *nfused,T3D *m2o)
{
   int i,j,me,mefus,lastp,c,k;
   double sum[21],vtmp[6];
   static T3D transf2befused[NB_MAX_ME];
   int ok[NB_MAX_ME],labels[NB_MAX_ME];

   /** Useful structures */
   /** transf2fused  contiendra les estimes  fusionner */
   for (i=0;i<NB_MAX_ME;i++) ok[i]=1;

   /** Prpare les flags de m2o, le rsultat de cette fonction  */
   t3dInit(m2o, T3D_U_MATRIX, T3D_ALLOW_CONVERSION);

   /** Nfused is the number of MEs we will fuse */
   *nfused=0;

   /** On passe en revue tous les estimateurs */
   for (me=0;me<set->nmes;me++)

     /** If we have a position for me at index and if it is fuseable */
     if ((set->me[me].p[index]!=NULL)&&(set->me[me].fuse)) {

       /** We add a pose to fuse */
       /** Cas ou l'estimateur est un estimateur absolu : segloc, locPost */
       if (set->me[me].p[index]->pos.class == POM_ABSOLUTE_ME) {

	 /** If the updater ME is absolute, get its pose and covariance */
	 /** On le met sous cette forme dans trans2befused */
	 t3dCopyIn(&transf2befused[*nfused],
		   &set->me[me].p[index]->pos.pos.abs.mainToOrigin);
	 labels[*nfused]=me;
	 /** Convert the pose to fuse in Euler form */
	 t3dConvertTo(T3D_EULER,&transf2befused[*nfused]);

#ifdef POM_DEBUG
	 fprintf(stderr,"Adding for fusion : loc (%d)\n",me);
	 dm_print(transf2befused[*nfused].euler.euler,1,6);
	 fprintf(stderr,"\n");
#endif
	 /** Compteur d'estimes  fusionner */
	 (*nfused)+=1;
	 /** Dernire estime */
	 mefus=me;

	 /** Pour l'instant on indique que l'on ne fusionne pas le me 0 */
	 set->me[0].fuse=0;
       }


       /** Cas des poses Relatives (xr4000...)  */
       /** Only for the estimation movements updated BEFORE or AT current index  */
       else if ((set->me[me].lastabs.date==100)
		/* Cas ou il n'a pas encore ete fusionn */
		||(pomDateIndex(set,set->me[me].lastabs.date) > index))
	 /* Cas ou la date de derniere fusion (par integration) prcde la date courante */
	 {
	   /** Else get the corresponding reference pose
	       and computes the covariance matrix in consequence
	       Removing the correlation terms
	       See note from Ph. Moutarlier : it can be generalized
	       as in the following lines */
	  /*  fprintf(stderr,"(last %d) ",set->me[me].lastabs.date); */
#ifdef POM_DEBUG
	   fprintf(stderr,"Adding for fusion : odo (%d) for index %d \n\n",
		   me,index);
#endif
	   /** A l'init */
	   /** On prend la pose telle quelle */
	   if (set->me[me].lastabs.date==100){
	     t3dCopyIn(&transf2befused[*nfused], &set->me[me].p[index]->pos.pos.abs.mainToOrigin);
	     /* printf("Initialisation de l'intgration de %s\n",set->me[me].name); */
	   } else {
	     /** Integrate from lastabs ! */
	     pomIntegrateBaseToOrigin(&transf2befused[*nfused],set,me,index);
	   }

	   /** Convertit en Euler */
	   t3dConvertTo(T3D_EULER,&transf2befused[*nfused]);
#ifdef POM_DEBUG
/* 	   dms_print(transf2befused[*nfused].euler.v,6,6); */
#endif
	   /** Incrmente le compteur de poses  fusionner */
	   labels[*nfused]=me;
	   (*nfused)+=1;
	   mefus=me;
#ifdef POM_DEBUG
	   fprintf(stderr," %d ",me);
#endif
	   if (set->log!=NULL) gzprintf(set->log,
					"Fusing (%d)\n",me );
	 }

       /** Other : here, it means that the odometry has already be integrated
	for a more recent date */
       else {
#ifdef POM_DEBUG
 	 fprintf(stderr,"Can't add odo (%d) for index %d - last MAJ :  (date %d)\n\n",
 		 me,index,set->me[me].lastabs.date);
#endif
	 /** Start integration from the last valid pose */
	 /** The last valid pose SUCH AS we also have a me pose */
	 lastp=pomMELastPos(set,0,index+1);
	 if (lastp>=0)
	   while (set->me[me].p[lastp]==NULL) {
	     lastp=pomMELastPos(set,0,lastp+1);
	 }

	 /** Check if we have one or not */
	 if (lastp<0) {
	   printf("This ME %d cannot be updated now (in %d)\n",me ,index);
	   if (set->log!=NULL) gzprintf(set->log,
					 "This ME %d cannot be updated now (in %d)\n",
					me ,index);
	 } else {
	   /* Update le B20 pour le me "me" */
	   /* Mmorise galement le mainToOrigin au moment de cet update */
	   /* cela est ncessaire pour un calcul propre des incertitudes */
	   pomSingleUpdateB2O(&set->me[0].p[lastp]->pos.pos.abs.mainToOrigin,
			      set, me, lastp);

	   /* Intgre par composition de basetoorigin sur les valeurs courantes de me*/
	   if (!pomIntegrateBaseToOrigin(&transf2befused[*nfused], set, me, index)) {
#ifdef POM_DEBUG
	     fprintf(stderr,"(last %d) ", set->me[me].lastabs.date);
	     fprintf(stderr," %d ",me);
#endif
	     /* rcupre cette pos een Euler */
	     t3dConvertTo(T3D_EULER,&transf2befused[*nfused]);
	     mefus=me;
	     labels[*nfused]=me;
	     (*nfused)+=1;
	   }
	 }
       }
     }

   /** Nothing to fuse */
   if ((*nfused)<1) {
     fprintf(stderr,"pomKalmanUpdateB2O: nothing to fuse\n");
     if (set->log!=NULL) gzprintf(set->log,
				  "pomKalmanUpdateB2O: nothing to fuse\n");
     return 1;
   }

   /** We are just integrating relative movement */
   if (((*nfused)==1)
       &&(set->me[mefus].p[index]->pos.class != POM_ABSOLUTE_ME))  {
     /** Then, don't fuse next things with me[0] */
     set->me[0].fuse=0;
   }


   /** FUSION */
   switch (set->method) {
   case POM_BAYES:
   case POM_KALMAN:
     /** Fusion */
     /** Case  we have only one estimator */
     if ((*nfused)==1) {
#ifdef POM_DEBUG
       fprintf(stderr,"Just taking one value\n");
#endif
       t3dConvertTo(T3D_EULER,&transf2befused[0]);
       t3dConvertTo(T3D_EULER,m2o);
       /** Just copy values and covariance */
       memcpy(m2o->euler.euler, transf2befused[0].euler.euler, 6*sizeof(double));
       memcpy(m2o->euler.v, transf2befused[0].euler.v, 21*sizeof(double));
     } else {

       /** Init sum */
       for (i=0;i<21;i++) sum[i]=0.0;
       t3dConvertTo(T3D_EULER,m2o);

       /** Init new value */
       for (i=0;i<6;i++) m2o->euler.euler[i]=0.0;

       /** Sum computation */
       /** sum = \sigma_{i=1}^{n} (V_i)^{-1} */
       for (i=0;i<*nfused;i++) {
	 t3dConvertTo(T3D_EULER,&transf2befused[i]);
	 /** To avoid problems while doing te innversion */
	 for (j=0;j<6;j++)
	   if (fabs(transf2befused[i].euler.v[j*(j+1)/2+j])<0.00000001)
	     transf2befused[i].euler.v[j*(j+1)/2+j]=1.0;
#ifdef POM_DEBUG
	 fprintf(stderr,"To be fused:\n");
	 fprintf(stderr,"Value:\n");
	 t3dFPrint(stderr, transf2befused[i]);
#endif
	 /* Inversion and summation */
	 pom_dms_mulscal_6x6(transf2befused[i].euler.v,1000.0);
	 if (check_cov_matrix(transf2befused[i].euler.v, set->log))
	   printf("Bad covariance matrix for [%d] ", labels[i]);
	 if (pom_dms_inv_6x6(transf2befused[i].euler.v, transf2befused[i].euler.v)) {
	   fprintf(stderr,"Can't do fusion with %s !\n",set->me[labels[i]].name);
	   if (set->log!=NULL) {
	     gzprintf(set->log,"Can't do fusion with %s !\n",set->me[labels[i]].name);
	     gzprintf(set->log,"Check this matrix !\n");
	     for(c=0,j=0;j<6;j++) {
	       gzprintf(set->log,"\t |");
	       for(k=0,c+=j;k<6;k++)
		 if (k>j) gzprintf(set->log,"         ");
		 else gzprintf(set->log,"%8.4f ", transf2befused[i].euler.v[c+k]);
	       gzprintf(set->log,"|\n");
	     }
	   }
	   ok[i]=0;
	 }
	 pom_dms_mulscal_6x6(transf2befused[i].euler.v,1000.0);
	 pom_dms_plus_6x6(sum,transf2befused[i].euler.v,sum);
       }

       /* Inversion */
       pom_dms_mulscal_6x6(sum,1000.0);
       check_cov_matrix(sum, set->log);
       if (pom_dms_inv_6x6(sum, m2o->euler.v)) {
	 /* What can i do in this case ????? XXXX */
	 return 1;
       }

       pom_dms_mulscal_6x6(m2o->euler.v,1000.0);


       /* Fuse all fuseable */
       for (i=0;i<*nfused;i++) if (ok[i]) {
	 /* Get the i transform in its euler form */
	 /* Ponderate it by the inverse of the covariance */
	 pom_dmsf_mult(transf2befused[i].euler.v,
		       transf2befused[i].euler.euler,
		       vtmp,6,6,6,1);
	 /* Adds it */
	 for(k=0;k<6;k++)
	   m2o->euler.euler[k] += vtmp[k];
       }
       /* And finally multiply by the sum matrix */
       pom_dmsf_mult(m2o->euler.v,m2o->euler.euler, vtmp, 6,6,6,1);
       memcpy(m2o->euler.euler, vtmp, 6*sizeof(double));
       pom_dms_mulscal_6x6(m2o->euler.v,(double)((*nfused)));
     }
     break;
   default:
     break;

   }

   /** To avoid problems while doing the inversion */
#define index_smatelt(i,j)   \
   ((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j))

   for (j=0;j<6;j++)
     if (fabs(m2o->euler.v[j*(j+1)/2+j]-1.0/ (double)(*nfused))<0.00000001) {
       for (i=0;i<6;i++) {
	 m2o->euler.v[index_smatelt(j,i)]=0.0;
	 m2o->euler.v[index_smatelt(i,j)]=0.0;
       }
       m2o->euler.euler[j]=0.0;
       fprintf(stderr, "warning, reset matrix\n");
     }
#undef index_smatelt

#ifdef POM_DEBUG
   fprintf(stderr,"Fusion result :\n");
   dm_print(m2o->euler.euler,1,6);
/*    printf("\n\n"); */
/*    dms_print(m2o->euler.v,6,6); */
#endif
   t3dConvertTo(T3D_U_MATRIX,m2o);


   /** update other's b2o */
   for(i=1;i<set->nmes;i++) {
     if (set->me[i].fuse &&
	 set->me[i].p[index] != NULL ) {
       /** No b2o update in the case of an only EM pose */
       if ((*nfused==1)&&(mefus==i)&&(set->me[i].p[index]->pos.class!=POM_ABSOLUTE_ME)) {
#ifdef POM_DEBUG
	 fprintf(stderr,"Estimator %d must not be updated in this case\n",i);
#endif
	 if (set->log!=NULL) gzprintf(set->log,
				       "Estimator %d must not be updated in this case\n",i);
       } else {
	 /** Normal update */
#ifdef POM_DEBUG
	 fprintf(stderr,"Update %d from Fusion \n",i);
#endif
	 if (set->log!=NULL) gzprintf(set->log,
				       "Update %d from Fusion \n",i);
	 pomSingleUpdateB2O(m2o,set,i,index);
       }
     }
   }

   return 0;
}



/**
 * Calcul de BaseToOrigin
 */

int
pomSingleUpdateB2O(T3D *m2o,PomMESet *set,int me,int index)
{

  set->me[me].lastabs.date = set->me[me].p[index]->pos.pos.abs.date;
  t3dCopyIn(&set->me[me].lastabs.mainToOrigin,
	    &set->me[me].p[index]->pos.pos.abs.mainToOrigin);

  /** Copy MainToOrigin */
  t3dCopyIn(&set->me[me].baseToOrigin,
	    &set->me[me].p[index]->pos.pos.abs.mainToOrigin);

  /** Inverts it */
  t3dInvert(&set->me[me].baseToOrigin);
  /** Compose it with with the updater ME pose
      The covariance matrix is also updated in the process */
  t3dComp(&set->me[me].baseToOrigin, m2o);
  t3dConvertTo(T3D_U_MATRIX,&set->me[me].baseToOrigin);

#ifdef POM_DEBUG
  printf("B2O for me %d\n",me);
/*   dms_print(set->me[me].baseToOrigin.matrix.v,6,6); */
#endif

  return 0;
}
