/*	$Id: pom.c 2007/08/24 21:28:37 mallet $ */

/*
 * Copyright (c) 2003-2007 LAAS/CNRS                  --  Mon Jul  3 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: pom.c 2007/08/24 21:28:37 mallet $";

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>

#include "t3d/t3d.h"

#include "libpom/lib.h"
#include "pomInt.h"

#define LOC_MAX(x,y) ((x>y)?(x):(y))

static int init=0;

/** Motion estimators that pom will consider */
static PomMESet meSet;

/** List of free pos structs to be used */
static PomMEList *freeList = NULL;

/******************************************************************************
*
*   angleLimit  -  Limiter la valeur d'un angle entre -PI et PI
*
*   Description:
*   Decremente de 2PI la valeur d'un angle, jusqu'a ce que la valeur de cet
*   angle se trouve entre -PI et PI ([-PI, PI])
*
*   Retourne: le resultat de la limitation
*/

static double
angleLimit(double angle)
{
  /* Test angle */
  if (fabs(angle) > 10000) {
    printf ("angleLimit: bad angle %f\n", angle);
    return 0;
  }

  /* Ajuster l'angle s'il est superieur a PI */
  while (angle > M_PI)
    angle -= 2.*M_PI;

  /* L'ajuster s'il est inferieur a -PI */
  while (angle < -M_PI)
    angle += 2.*M_PI;

  /* Retourner l'angle ajuste */
  return (angle);
}


#if 0
int
pomVBaseToVZero(int ind)
{
  int lastvb,i;
  T3D *tmp,*tmp2;

  /* Last index before 'date' where we have refme available */
  lastvb  = pomMELastPos(&meSet, meSet.refMe, ind);

  if (lastvb <0)
    return 1;

  /* on rcupre le baseToOrigin de refMe */
  t3dConvertTo(T3D_U_MATRIX,&meSet.me[meSet.refMe].baseToOrigin);

  /* A new transf */
  tmp = t3dNew(T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
  tmp2 = t3dNew(T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
  t3dCompIn(tmp,&meSet.me[meSet.refMe].offset,&meSet.me[meSet.refMe].baseToOrigin);
  t3dInvertIn(tmp2,&meSet.originToZero);
  t3dComp(tmp,tmp2);
  t3dDelete(tmp2);

  /* Just keep the rotation */
  t3dConvertTo(T3D_EULER,tmp);
  tmp->euler.euler[3]=0.0;  tmp->euler.euler[4]=0.0; tmp->euler.euler[5]=0.0;

  /* Apply the rotation form of B2O */
  t3dCompIn(&meSet.me[0].p[lastvb]->pos.velocity.VOrigin,
	    &meSet.me[meSet.refMe].p[lastvb]->pos.velocity.VOrigin,
	    tmp);
  t3dConvertTo(T3D_EULER,&meSet.me[0].p[lastvb]->pos.velocity.VOrigin);
  t3dConvertTo(T3D_EULER,&meSet.me[meSet.refMe].p[lastvb]->pos.velocity.VOrigin);
  /* For angular motion, velocities remain the same ... */
  /* To BE  CHANGED FOR GENERALIZATION IN 3D MOTION */
  for (i=0;i<3;i++)
    meSet.me[0].p[lastvb]->pos.velocity.VOrigin.euler.euler[i]=
      meSet.me[meSet.refMe].p[lastvb]->pos.velocity.VOrigin.euler.euler[i];

  if ((fabs(meSet.me[meSet.refMe].p[lastvb]->pos.velocity.VOrigin.euler.euler[3])>0.05)||(fabs(meSet.me[meSet.refMe].p[lastvb]->pos.velocity.VOrigin.euler.euler[4])>0.05)) {
    dm_print(meSet.me[0].p[lastvb]->pos.velocity.VOrigin.euler.euler,1,6);
  }
  t3dDelete(tmp);

  return 1;
}
#endif

/*
 * Gives the displacement measured
 * by the ref Estimator between two poses
 */
T3D *
pomGetRefDisplacement(int date1, int date2)
{
  return ((T3D *)pomRefDisplacement(&meSet,date1,date2));
}


/**
 * pomTick ----------------------------------------------
 *
 * Increment the current pom's date.
 */

int
pomTick()
{
   int i, j, k;
   int index,tmp;

   /** Increment the current pom's date */
   meSet.currentTime++;

   /** find the smallest index that includes two positions of every me */

   index = 0;

   for(i=1;i<meSet.nmes;i++) { /** me #0 is the fused position */


     /** look for the two last pos */
     for(j=tmp=k=0;j<meSet.length;j++)
       if (meSet.me[i].p[j] != NULL) {
	 k++;
	 if (k>0 && k<3) { /** keep at most 2 pos */
	   if (tmp < j) tmp = j;
	 }
       }
     if (tmp>=meSet.length-1) {
       fprintf(stderr,"Desactivating ME %d \n",i);
       if (meSet.log!=NULL) gzprintf(meSet.log,
				     "Desactivating ME %d \n",i);

       meSet.me[i].fuse=0;
     }

     if (meSet.log!=NULL) gzprintf(meSet.log,
				   "Index to get two poses from %d : %d \n",i,tmp);
     /** do not consider display-only estimators */
     if (!meSet.me[i].fuse) continue;

     /** look for the two last pos */
     for(j=k=0;j<meSet.length;j++)
       if (meSet.me[i].p[j] != NULL) {
	 k++;
	 if (k>0 && k<3) { /** keep at most 2 pos */
	   if (index < j) index = j;
	 }
       }
   }

   /** resize meSet if needed */
   /* if (pomMESetResizeL(&meSet, meSet.currentTime - (index+1))) return -1; */

   /** move data */
   for(i=0;i<meSet.nmes;i++) {
     /** The last position in the history */
      if (meSet.me[i].p[meSet.length-1] != NULL) {
	/** If it is not null we recuperate it into the list, faut pas gacher */
	meSet.me[i].p[meSet.length-1]->next = freeList;
	freeList = meSet.me[i].p[meSet.length-1];
      }
      /** Shifts everything towards left */
      memmove(meSet.me[i].p+1, meSet.me[i].p,
	      (meSet.length-1)*sizeof(PomMEList *));
      /** Keeps a place for the new position, if it arrives */
      meSet.me[i].p[0] = NULL;
   }

   /** If this tick occurs while an update/fuse is being performed ... */
   meSet.fusionIndex++;

   if (meSet.fusionIndex>=POM_LE_PACKET_SIZE)
     meSet.fusionIndex=POM_LE_PACKET_SIZE-1;

   return meSet.currentTime;
}

int
pomGetTick()
{
  int date = meSet.currentTime;
  return date;
}


/**
 * pomLog ---------------------------------------------------------------
 *
 * start (or stop if "file" is NULL) data logging.
 */

gzFile *
pomLog(gzFile *file)
{
   gzFile *tmp;

   tmp = meSet.log;
   meSet.log = file;
   return tmp;
}


/**
 * pomSetZero -----------------------------------------------------------
 *
 * Compute the originToZero offset so that the position becomes equals to
 * "mainToOrigin".
 */

int
pomSetZero(T3D *mainToOrigin)
{
   int best;

   t3dConvertTo(T3D_U_MATRIX, mainToOrigin);

   /** get the last fused pos */
   best = pomMELastPos(&meSet, 0, 0);
   if (best < 0) {
     /** if there is no position for now, we're at origin */
     meSet.originToZero = *mainToOrigin;
     meSet.originToZero.flags = T3D_ALLOW_CONVERSION;
     return 0;
   }

   t3dCopyIn(&meSet.originToZero,
	     &meSet.me[0].p[best]->pos.pos.abs.mainToOrigin);
   t3dInvert(&meSet.originToZero);
   t3dComp(&meSet.originToZero, mainToOrigin);

   /*GUILLO: small error that took 3 hours to find*/
   /* for two person, that makes 6 hours...*/
   t3dInvert(&meSet.originToZero);

   return 0;
}


/**
 * pomGetPos ------------------------------------------------------------
 *
 * Return the last pos for me (debug and display only)
 */

int
pomGetPos(int me, T3D *pos)
{
   int last;

   last = pomMELastPos(&meSet, me, 0);
   if (last < 0) return -1;

   t3dCopyIn(pos, &meSet.me[me].p[last]->pos.pos.abs.mainToOrigin);
   t3dComp(pos, &meSet.originToZero);

   return 0;
}


/**
 * pomSetRefME ----------------------------------------------------------
 *
 * Set the reference estimator (baseToOrigin source)
 */

int
pomSetRefME(int me)
{
   if (me < 1 || me >=meSet.nmes) return -1;

   if (me == 0) {
      meSet.me[0].baseToOrigin = t3dUMatrixIdentity;
      meSet.me[0].baseToOrigin.flags = T3D_ALLOW_CONVERSION;
   }

   meSet.refMe = me;
   return 0;
}

int
pomSetFusionMethod(PomFusionMethod fusionMethod)
{
  meSet.method=fusionMethod;
  return 0 ;
}

int
pomInit(int fusionMethod)
{
  if (!init) {
    meSet.nmes=0;
    meSet.currentTime=0;
    meSet.fusionIndex=-1;
    meSet.me=NULL;
    meSet.length=0;
    meSet.packets=0;
    meSet.refMe=0;
    meSet.log=NULL;
    meSet.method=fusionMethod;
    init = 1;
  }
  return 0;
}

/**
 * pomReset -------------------------------------------------------------
 *
 * Erase All ME and reinitialize meSet.
 * Returns -1 on error.
 */
int
pomReset() {
  int i;

  for(i=0;i<meSet.packets;i++) {
    /* free positions history */
    free(meSet.me[i].p);
  }
  free(meSet.me);
  freeList = NULL;
  init = 0;
  pomInit(meSet.method);
  return 0;
}

/**
 * pomMECreate ----------------------------------------------------------
 *
 * Create and register a new motion estimator (ME) and return an index that
 * must be used as a reference to this ME.
 * Returns -1 on error.
 */

int
pomMECreate(char *name, int fuse)
{
   int i, index = -1;

   /** check if "name" already exist... */
   for(i=1; i<meSet.nmes; i++) if (!strcmp(name, meSet.me[i].name)) {
      index = i;
      errno = EINVAL;
      return -1;
   }

   /** ...or create me if needed */
   if (index < 0) {
      /** check size of meSet */
      if (pomMESetResizeN(&meSet, meSet.nmes+1)) return -1;

      /** create the new ME */
      index = meSet.nmes++;

      meSet.me[index].isOffset = 0;
      meSet.me[index].name = calloc(strlen(name)+1,sizeof(char));
      strcpy(meSet.me[index].name, name);

      for(i=0;i<meSet.length;i++) meSet.me[index].p[i] = NULL;

      meSet.me[index].baseToOrigin = t3dUMatrixIdentity;
      meSet.me[index].baseToOrigin.flags = T3D_ALLOW_CONVERSION;

      t3dSetOk(&meSet.me[index].baseToOrigin);
      /** Initialize the lastabs element */
      meSet.me[index].lastabs.date=100;
   }

   meSet.me[index].fuse = fuse;
   return index;
}


/**
 * pomMERemove ----------------------------------------------------------
 *
 * Remove a position estimator from list. Actually, we juste reset the
 * "fuse" flags to 0 (the position we still be read but it won't be
 * included in fusion).
 */

int
pomMERemove(int me)
{
   if (me < 1 || me >= meSet.nmes) {
      fprintf(stderr, "pom: bad motion estimator\n");
      errno = EINVAL;
      return -1;
   }

   meSet.me[me].fuse = 0;
   return 0;
}

/**
 * pomMEDeltaUpdate ----------------------------------------------------------
 *
 * Update a Delta motion estimator with a displacement.
 */
int
pomMEDeltaUpdate(int me, LibPomMEPos *pos)
{
   PomMEList *lpos;
   int last, index, date;

   printf ("pomMEDeltaUpdate: too dangerous : should disapear\n");

#ifdef POM_DEBUG
   fprintf(stderr,"update me %d\n",me);
#endif
   if (me < 1 || me >= meSet.nmes) {
      fprintf(stderr, "pom: bad motion estimator\n");
      errno = EINVAL;
      return -1;
   }

   switch(pos->posType) {
   case POM_ABSOLUTE_ME:	return -1;
   case POM_DELTA_ME:	date = pos->pos.delta.date2; break;
   default:
     fprintf(stderr, "pom: bad position for %s\n", meSet.me[me].name);
     errno = EINVAL;
     return -1;
   }

   if (date < 0) {
     /** the position is not valid: forget it */
     return 0;
   }

   /** check history length */
   if (pomMESetResizeL(&meSet, date)) return -1;

   /** get a pos struct */
   if (freeList != NULL) {
     /** Update freeList not to take lpos into account */
     lpos = freeList;
     freeList = freeList->next;
   } else {
     /** Else allocates a new pose structure */
     lpos = calloc(1,sizeof(PomMEList));
     fprintf(stderr,"Allocation in DeltaUpdate !\n");
     if (lpos == NULL) { errno = ENOMEM; return -1; }
   }
   lpos->pos = *pos;

   t3dConvertTo(T3D_U_MATRIX, &lpos->pos.pos.abs.mainToOrigin);

   /** get pos index in history corresponding to date */
   index = pomDateIndex(&meSet, date);
   if (index < 0) {
     fputs("pom: date out of range for ", stderr);
     fputs(meSet.me[me].name, stderr);
     fputs("\n", stderr);
     return -1;
   }

   /** compute offset if this is the first time we get a pos */
   if (!meSet.me[me].isOffset) {
     meSet.me[me].offset = lpos->pos.pos.abs.mainToOrigin;
     t3dInvert(&meSet.me[me].offset);
     /** Gets the last time fusion was realized */
     last = pomMELastPos(&meSet, 0, 0);
     if (last >= 0) {
       t3dComp(&meSet.me[me].offset,
	       &meSet.me[0].p[last]->pos.pos.abs.mainToOrigin);
     }
     t3dConvertTo(T3D_EULER,&meSet.me[me].offset);
     meSet.me[me].isOffset = 1;
   }

   /** concatenate delta position to the last pos */
   last = pomDateIndex(&meSet, lpos->pos.pos.delta.date2);

   last = pomMELastPos(&meSet, me, last);
   if (last >= 0) {
     t3dComp(&lpos->pos.pos.delta.main,
	     &meSet.me[me].p[last]->pos.pos.abs.mainToOrigin);
   }
   lpos->pos.class = POM_DELTA_ME;
   lpos->pos.posType = POM_ABSOLUTE_ME;
   lpos->pos.pos.abs.date = date;

   /** register pos struct */
   if (meSet.me[me].p[index] != NULL) {
/*      fprintf(stderr, "pom: %s position overriden\n", meSet.me[me].name); */
     meSet.me[me].p[index]->next = freeList;
     freeList = meSet.me[me].p[index];
   }


   meSet.me[me].p[index] = lpos;
   meSet.me[me].fuse=1;
   if (meSet.fusionIndex < index) meSet.fusionIndex = index;
   return 0;
}


/*----------------------------------------------------------------------*/
/* return 0 if ok */
static int
testJump(T3D *new, T3D *last,
	 double seuilCoord, double seuilAngle,
	 double coeffMultCoord, double coeffMultAngle)
{
  double ut,uph,ups,ux,uy,uz;
  double et,eph,eps,ex,ey,ez;

  ut = POM_3SIGMA(last->euler.v[0]);
  uph = POM_3SIGMA(last->euler.v[2]);
  ups = POM_3SIGMA(last->euler.v[5]);
  ux = POM_3SIGMA(last->euler.v[9]);
  uy = POM_3SIGMA(last->euler.v[14]);
  uz = POM_3SIGMA(last->euler.v[20]);

  et = angleLimit(new->euler.euler[0]-last->euler.euler[0]);
  eph = angleLimit(new->euler.euler[1]-last->euler.euler[1]);
  eps = angleLimit(new->euler.euler[2]-last->euler.euler[2]);
  ex = new->euler.euler[3] - last->euler.euler[3];
  ey = new->euler.euler[4] - last->euler.euler[4];
  ez = new->euler.euler[5] - last->euler.euler[5];

  if (fabs(ex) > (ux*coeffMultCoord + seuilCoord) ||
      fabs(ey) > (uy*coeffMultCoord + seuilCoord) ||
      fabs(ez) > (uz*coeffMultCoord + seuilCoord) ||
      fabs(et) > (ut*coeffMultAngle + seuilAngle) ||
      fabs(eph) > (uph*coeffMultAngle + seuilAngle) ||
      fabs(eps) > (ups*coeffMultAngle + seuilAngle))
    return -1;

  return 0;
}

/**
 * pomMEUpdate ----------------------------------------------------------
 *
 * Update a motion estimator with a new position.
 */

int
pomMEUpdate(int me, LibPomMEPos *pos)
{
   PomMEList *lpos;
   int last, index, date,i;
#ifdef POM_DEBUG
   fprintf(stderr,"update me %d\n",me);
#endif
   if (meSet.log!=NULL) gzprintf(meSet.log,
				 "update me %d\n",me);
   if (me < 1 || me >= meSet.nmes) {
      fprintf(stderr, "pom: bad motion estimator\n");
      errno = EINVAL;
      return -1;
   }

   switch(pos->posType) {
      case POM_ABSOLUTE_ME:	date = pos->pos.abs.date; break;
      case POM_DELTA_ME:
	printf ("pomMEUpdate: POM_DELTA_ME too dangerous : should disapear\n");
	date = pos->pos.delta.date2; break;
      default:
	 fprintf(stderr, "pom: bad position for %s\n", meSet.me[me].name);
	 errno = EINVAL;
	 return -1;
   }

   if (date < 0) {
     /** the position is not valid: forget it */
     return 0;
   }



   if (meSet.currentTime - date > meSet.length) {
     /** if this is the first time we get a pos, we suppose that the robot has not moved !  */
     /** So we give it a dummy date */
     if (!meSet.me[me].isOffset) {
       date = meSet.currentTime - 1;
     } else {
       printf("pomMEUpdate : me %d : memory full (current time %d  - date required %d > length %d)\n",
	      me, meSet.currentTime, date, meSet.length);
       if (meSet.log!=NULL) gzprintf(meSet.log,
				   "pomMEUpdate : I can't take this pose into account for memory reasons\n");
       return 0;
     }
   }

   /** get a pos struct */
   if (freeList != NULL) {
     /** Update freeList not to take lpos into account */
     lpos = freeList;
     freeList = freeList->next;
   } else {
     /** Else allocates a new pose structure */
     lpos = calloc(1,sizeof(PomMEList));
     if (lpos == NULL) { errno = ENOMEM; return -1; }
   }
   lpos->pos = *pos;

   t3dConvertTo(T3D_U_MATRIX, &lpos->pos.pos.abs.mainToOrigin);

   /** get pos index in history corresponding to date */
   index = pomDateIndex(&meSet, date);
   if (index < 0) {
      fputs("pom: date out of range for ", stderr);
      fputs(meSet.me[me].name, stderr);
      fputs("\n", stderr);
      return -1;
   }

   /** compute offset if this is the first time we get a pos */
   if (!meSet.me[me].isOffset) {

     meSet.me[me].offset = lpos->pos.pos.abs.mainToOrigin;
     t3dInvert(&meSet.me[me].offset);

      /** Gets the last time fusion was realized */
     last = pomMELastPos(&meSet, 0, index);
     if (last >= 0) {
       t3dComp(&meSet.me[me].offset,
	       &meSet.me[0].p[last]->pos.pos.abs.mainToOrigin);
     }

     t3dConvertTo(T3D_EULER,&meSet.me[me].offset);
     for (i=0;i<21;i++) meSet.me[me].offset.euler.v[i]=0.0;
     if (meSet.log!=NULL) gzprintf(meSet.log,
				   "Offset set for me %d, equal to me[0] at index %d \n",
				   me,last);
     printf("Offset set for me %d, equal to me[0] at index %d \n",me,last);
     meSet.me[me].isOffset = 1;
   }



   switch (pos->posType) {
      case POM_ABSOLUTE_ME:
	/** add offset to the position */

	t3dComp(&lpos->pos.pos.abs.mainToOrigin, &meSet.me[me].offset);
	t3dConvertTo(T3D_EULER,&lpos->pos.pos.abs.mainToOrigin);
	/* FOR DEBUG */
	if (pos->class==POM_ABSOLUTE_ME) {
	  last = pomMELastPos(&meSet, 0, index);
	  if (last>0) {
	    t3dConvertTo(T3D_EULER,&meSet.me[0].p[last]->pos.pos.abs.mainToOrigin);

	    /* test si la nouvelle position donnee par l'estimateur est
	     * coherente avec la derniere position fusionee. XXXXXXXXX
	     * Attention: devrait etre compare a la position fusionnee
	     * corrspondant a l'instant d'aquisition. prevoir aussi des
	     * seuils non harcodes (du moins les mettre dans pomConst) */
	    if (testJump(&lpos->pos.pos.abs.mainToOrigin,
			 &meSet.me[0].p[last]->pos.pos.abs.mainToOrigin,
			 0.5,0.2,1.2,1.2) == -1) {
	      fprintf(stderr,
			 "PBM IN YOUR POSE !!\n Direct ME pose ");
	      for (i=0;i<6;i++)
		  fprintf(stderr,"%5.3f ",
			   pos->pos.abs.mainToOrigin.euler.euler[i]);
	      fprintf(stderr,"PBM IN YOUR POSE !!\n Pose you enter = ");
	      for (i=0;i<6;i++)
		  fprintf(stderr,"%5.3f ",
			   lpos->pos.pos.abs.mainToOrigin.euler.euler[i]);
	      fprintf(stderr,
			 "PBM IN YOUR POSE !!\n Pose current (%d) = ",last);
	      for (i=0;i<6;i++)
		  fprintf(stderr,"%5.3f ",
			   meSet.me[0].p[last]->pos.pos.abs.mainToOrigin.euler.euler[i]);
	      if (meSet.log!=NULL) {
		gzprintf(meSet.log,
			 "PBM IN YOUR POSE !!\n Pose you enter = ");
		for (i=0;i<6;i++)
		  gzprintf(meSet.log,"%f ",
			   lpos->pos.pos.abs.mainToOrigin.euler.euler[i]);
		gzprintf(meSet.log,
			 "PBM IN YOUR POSE !!\n Pose current (%d) = ",last);
		for (i=0;i<6;i++)
		  gzprintf(meSet.log,"%f ",
			   meSet.me[0].p[last]->pos.pos.abs.mainToOrigin.euler.euler[i]);

	      }

	      lpos->next = freeList;
	      freeList = lpos;

	      return -1;
	    }
	  }
	}
	break;
	 /**
	  * NOT USED !
	  * Too dangerous
	  */

      case POM_DELTA_ME:
	printf ("pomMEDeltaUpdate: POM_DELTA_ME too dangerous : should not be used\n");
	 /** concatenate delta position to the corresponding fused pos */
	 last = pomDateIndex(&meSet, lpos->pos.pos.delta.date1);
	 last = pomMELastPos(&meSet, 0, last);
	 if (last >= 0) {
	    t3dComp(&lpos->pos.pos.delta.main,
		    &meSet.me[0].p[last]->pos.pos.abs.mainToOrigin);
	 }
	 lpos->pos.class = POM_DELTA_ME;
	 lpos->pos.posType = POM_ABSOLUTE_ME;
	 lpos->pos.pos.abs.date = date;
	 break;
   }

   /** register pos struct */
   if (meSet.me[me].p[index] != NULL) {
      meSet.me[me].p[index]->next = freeList;
      freeList = meSet.me[me].p[index];
   }

   meSet.me[me].p[index] = lpos;
   meSet.me[me].fuse=1;
   if (meSet.fusionIndex < index) meSet.fusionIndex = index;
   t3dSetOk(&meSet.me[me].p[index]->pos.pos.abs.mainToOrigin);
   t3dConvertTo(T3D_EULER,&meSet.me[me].p[index]->pos.pos.abs.mainToOrigin);
#ifdef POM_DEBUG
   fprintf(stderr,"update me %d index %d fusionindex %d\n",me,index,meSet.fusionIndex);
   dm_print(meSet.me[me].p[index]->pos.pos.abs.mainToOrigin.euler.euler,1,6);
/*    fprintf(stderr,"\n---\n"); */
/*    dms_print(meSet.me[me].p[index]->pos.pos.abs.mainToOrigin.euler.v,6,6); */
#endif
   if (meSet.log!=NULL) gzprintf(meSet.log,
				 "update me %d index %d fusionindex %d\n",
				 me,index,meSet.fusionIndex);
   return 0;
}


/**
 * pomClass --------------------------------------------------------------
 *
 *
 */

PomMEClass
pomClass(int me)
{
  return pomMEClass(&meSet,me);
}

/**
 * pomFuse --------------------------------------------------------------
 *
 * Compute the best position estimate
 */

int
pomFuse(LibPomPos *pos)
{
   PomMEList *lpos;
   int best=0,dbest=0, last,nfused;
   T3D tmp,*tmp2=t3dNew(T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
   int lastodo;

   t3dInit(&pos->mainToBase, T3D_U_BRYAN, T3D_ALLOW_CONVERSION);
   t3dInit(&pos->mainToOrigin, T3D_U_BRYAN, T3D_ALLOW_CONVERSION);
   t3dInit(&pos->VLocal, T3D_U_BRYAN, T3D_ALLOW_CONVERSION);

   /**
    * Looks back in history...
    */
   while(meSet.fusionIndex >= 0) {

/*      printf ("pomFuse meSet.fusionIndex = %d\n", meSet.fusionIndex); */

     /** Chooses best position for the current index
	 (will be used  for POM_MAX_CONF only) */
     best = pomBestME(&meSet, meSet.fusionIndex);

/*      printf ("pomFuse pomBestME = %d\n", best); */

     /** if there is no availabe position, do not fuse */
     if (best < 0) { meSet.fusionIndex--; continue; }

     if (meSet.log!=NULL) gzprintf(meSet.log,
				   "\n****\nBest at index %d : ME%d\n",
				   meSet.fusionIndex,best);

     /** delete old position */
     if (meSet.me[0].p[meSet.fusionIndex] != NULL) {
       meSet.me[0].p[meSet.fusionIndex]->next = freeList;
       freeList = meSet.me[0].p[meSet.fusionIndex];
     }

     /** get a pos struct */
     if (freeList != NULL) {
       lpos = freeList;
       freeList = freeList->next;
     } else {

       lpos = calloc(1,sizeof(PomMEList));
       if (lpos == NULL) {
	 errno = ENOMEM;
	 t3dDelete(tmp2);
	 return -1;
       }
     }

     switch (meSet.method) {
     case POM_MAX_CONF:

       /** update the base to origin matrix of other estimators */
       /** and gives the current estimate of covariance */
       if (pomUpdateB2O(&meSet, best, meSet.fusionIndex,&tmp)) {
	 fputs("pom: internal error\n", stderr);
	 if (freeList==NULL) free(lpos);
	 return -1;

       }

       /** if we are not the ref me, we must compute the mainToBase */
       if (meSet.refMe > 0 && best != meSet.refMe) {
	 /** Gets the last index in which refMe has been updated */
	 /** If the refMe is the quickest one : last = meSet.fusionIndex */
	 last = pomMELastPos(&meSet, meSet.refMe, meSet.fusionIndex);
	 if (last < 0) {
	   printf("PBM : at baseToOrigin\n");
	   /** Init */
	   meSet.me[0].baseToOrigin = t3dUMatrixIdentity;
	   meSet.me[0].baseToOrigin.flags = T3D_ALLOW_CONVERSION;
	 }
	 lpos->pos = meSet.me[best].p[meSet.fusionIndex]->pos;
       } else /** the refMe is doing the update */ {
	 lpos->pos = meSet.me[best].p[meSet.fusionIndex]->pos;
	 t3dComp(&lpos->pos.pos.abs.mainToOrigin, &meSet.me[meSet.refMe].baseToOrigin);
	 /** So BaseToOrigin does not change for ME 0 */
       }

       break;

     case POM_BAYES:
     case POM_KALMAN:
#ifdef POM_DEBUG
       fprintf(stderr,"\n---------- Fusion using cov\n");
#endif
       /** update the base to origin matrix of other estimators */
       /** and gives the current estimate of covariance */
       if (pomKalmanUpdateB2O(&meSet, meSet.fusionIndex, &nfused, &tmp)) {
	 fputs("pom: internal error\n", stderr);
	 if (freeList==NULL) free(lpos);
	 t3dDelete(tmp2);
	 return -1;
       }

       if (nfused > 1)
	 if (meSet.log!=NULL) gzprintf(meSet.log, "FUSION DONE\n");

       /** Gets the last index in which refMe has been updated */
       /** If the refMe is the quickest one : last = meSet.fusionIndex */
       last = pomMELastPos(&meSet, meSet.refMe, meSet.fusionIndex);
       meSet.me[0].baseToOrigin = t3dUMatrixIdentity;
       meSet.me[0].baseToOrigin.flags = T3D_ALLOW_CONVERSION;
       lpos->pos.pos.abs.mainToOrigin = tmp;

       break;
     default:
       fprintf(stderr,"No such a fusion mode\n");
       if (freeList==NULL) free(lpos);
       t3dDelete(tmp2);
       return -1;
     } /* END CASE */


     /** store ref position and crushes the previous values in me[0] */
     meSet.me[0].p[meSet.fusionIndex] = lpos;
#ifdef POM_DEBUG
     fprintf(stderr,"Updating fusion index %d\n",meSet.fusionIndex);
#endif
     if (meSet.log!=NULL) gzprintf(meSet.log,
				   "Updating 0 at fusion index %d\n",
				   meSet.fusionIndex);
     meSet.me[0].p[meSet.fusionIndex]->pos.class=POM_ABSOLUTE_ME;

     /** Goes further in the history */
     meSet.fusionIndex--;

   }

   /** Get the last fused pos */
   /** Be careful, best is an index, here */
   best = pomMELastPos(&meSet, 0, 0);

#ifdef POM_DEBUG
   printf ("pomFuse : best = %d\n", best);
#endif

   if (best < 0) {
     /** Init , at the origin : mainToBase = id  */
     pos->mainToBase = t3dUMatrixIdentity;
     pos->mainToBase.flags = T3D_ALLOW_CONVERSION;
     /** Pose = OriginToZero */
     t3dCopyIn(&pos->mainToOrigin, &meSet.originToZero);
     /** Init VLocal */
     pos->VLocal = t3dUMatrixIdentity;
     pos->VLocal.flags = T3D_ALLOW_CONVERSION;
   } else {
     /** MaintoBase : the last odo pose */

     lastodo = pomMELastPos(&meSet,meSet.refMe,best);
     if (lastodo<0) {
       printf("No odo available\n");
     }
     else {
       pos->mainToBase = meSet.me[meSet.refMe].p[lastodo]->pos.pos.abs.mainToOrigin;
       /* This pose HAS TO be composed by an odometric offset
	that allows us to work in the same frame as xr4000
	(different from the one of the pose at which we started pom) */
       t3dInvertIn(tmp2,&meSet.me[meSet.refMe].offset);
       t3dComp(&pos->mainToBase,tmp2);


       /** Composed with OriginToZero */
       t3dInvertIn(tmp2,&meSet.originToZero);
       t3dCompIn(&pos->mainToOrigin,
		 &meSet.me[0].p[best]->pos.pos.abs.mainToOrigin,
		 tmp2);
     }


     /* On va chercher l'index correspondant
	a la pose odo la plus fraiche  (dbest) */
     dbest = pomMELastPos(&meSet,meSet.refMe,0);
     if (dbest>0) {
       /* Recopie de la vitesse locale */
       t3dCopyIn(&pos->VLocal,&meSet.me[meSet.refMe].p[dbest]->pos.velocity.VLocal);
     }

   }
   /** Tag by date */
   pos->date = meSet.currentTime;
   /** log to file */
   if ((best>0)&&(meSet.log != NULL))
     if (pomWrite(&meSet, best)) fputs("pom: log failed\n", stderr);

   /** Free tmp2 */
   t3dDelete(tmp2);
   return 0;
}

int
pomSetPosFromME(int me)
{
  int best,best0,i;
  T3D *tmp=t3dNew(T3D_U_MATRIX, T3D_ALLOW_CONVERSION);

  /* Get last pose from me "me"
   */
  best = pomMELastPos(&meSet, me, 0);
  if (best < 0) {
    fprintf(stderr,"Sorry, no pose from me %d available\n",me);
    return 1;
  }

  /* Get last fused pose
   */
  best0 = pomMELastPos(&meSet,0, 0);
  if (best0 < 0) {
    fprintf(stderr,"Sorry, no pose from me 0 available\n");
    return 1;
  }

  /* Inverts the best me pose (in pom_start frame)
   */
  t3dInvertIn(tmp,&meSet.me[me].p[best]->pos.pos.abs.mainToOrigin);

  /* Compose with the offset */
  t3dCompIn(&meSet.originToZero,&meSet.me[me].offset,tmp);

  /* Gets the transform between the two estimates
   */
  t3dComp(&meSet.originToZero,&meSet.me[0].p[best0]->pos.pos.abs.mainToOrigin);

  /* Convertit en euler
   */
  t3dConvertTo(T3D_EULER,&meSet.originToZero);

  /* For reasons of convenirnce (this is just an offset)
   * set its covariance to ZERO
   */
  for (i=0;i<21;i++) meSet.originToZero.euler.v[i]=0.0;


  t3dDelete(tmp);
  return 0;
}
