/*	$Id: refdiff.c 2007/08/22 19:12:05 mallet $ */

/*
 * Copyright (c) 2003-2007 LAAS/CNRS                  --  Thu May 22 2003
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: refdiff.c 2007/08/22 19:12:05 mallet $";

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "t3d/t3d.h"
#include "libpom/lib.h"

#include "pomInt.h"

/**
 * Gives the displacement of me ref odo between dates date1 and date2
 * and computes the exact value of the displacement uncertainty
 */
T3D *
pomRefDisplacement(PomMESet *set, int date1, int date2)
{
  T3D *displ,*tmp;
  int index1=pomDateIndex(set,date1),index2=pomDateIndex(set,date2);
  double Joo[36]; /** To simplify o : mainToOrigin^i and g : mainToOrigin at k+1 */
  double Jgt[36];
  double Jgo[36];
  double Jto[36],JRa[9],JRb[9],JRg[9],Jgtr[36];
  double dx,dy,dz,omega,rx[3],ry[3],rz[3],t[16];
  double m1[36],m2[36],m3[36],m4[36];
  int i,j,k;

  if ((index1<0)||(index2<0))
    {
      fprintf(stderr,"POM - pomRefDisplacement : Bad date values");
      return NULL;
    }
  /** Two new transformations */
  displ=t3dNew(T3D_U_MATRIX, T3D_ALLOW_CONVERSION);
  tmp  =t3dNew(T3D_U_MATRIX, T3D_ALLOW_CONVERSION);

  /** Le dplacement de l'odo */
  t3dInvertIn(tmp,&set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin);
  t3dCompIn(displ,
	    &set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin,
	    tmp);

  /** Conversion into Euler transforms, first */
  t3dConvertTo(T3D_EULER,&set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin);
  t3dConvertTo(T3D_EULER,&set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin);
  t3dConvertTo(T3D_EULER,tmp);
  pomMatrixOfEuler(tmp->euler.euler,t);
  /** Covariance */
  /** Get  Jgt  */
  pomDerivativeOfEuler(tmp->euler.euler, JRa,JRb,JRg);

  pom_dm_mult(JRa,set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler+3,rx,3,3,3,1);
  pom_dm_mult(JRb,set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler+3,ry,3,3,3,1);
  pom_dm_mult(JRg,set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler+3,rz,3,3,3,1);
  pom_dm_fill(Jgt,6,6,
	  1.0  ,0.0  ,0.0  ,0.0,0.0,0.0,
	  0.0  ,1.0  ,0.0  ,0.0,0.0,0.0,
	  0.0  ,0.0  ,1.0  ,0.0,0.0,0.0,
	  rx[0],ry[0],rz[0],1.0,0.0,0.0,
	  rx[1],ry[1],rz[1],0.0,1.0,0.0,
	  rx[2],ry[2],rz[2],0.0,0.0,1.0);

  /** Get  Jgo' */
  pom_dm_fill(Jgo,6,6,
	  1.0,0.0,0.0,0.0,0.0,0.0,
	  0.0,1.0,0.0,0.0,0.0,0.0,
	  0.0,0.0,1.0,0.0,0.0,0.0,
	  0.0,0.0,0.0,t[0],t[1],t[2] ,
	  0.0,0.0,0.0,t[4],t[5],t[6] ,
	  0.0,0.0,0.0,t[8],t[9],t[10]);


  /** Get  Jo'o */
  dx=set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler[3]-
    set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler[3];
  dy=set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler[4]-
    set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler[4];
  dz=set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler[5]-
    set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler[5];

  omega = 0.5*(set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.euler[0]+
	       set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler[0]);
  pom_dm_fill(Joo,6,6,
	  1.0,0.0                         ,0.0,0.0,0.0,0.0,
	  0.0,1.0                         ,0.0,0.0,0.0,0.0,
	  0.0,0.0                         ,1.0,0.0,0.0,0.0,
	  -dy,dz*cos(omega)               ,0.0,1.0,0.0,0.0,
	  dx ,dz*sin(omega)               ,0.0,0.0,1.0,0.0,
	  0.0,-dx*cos(omega)-dy*sin(omega),0.0,0.0,0.0,1.0);

  /** Get  Jto   */
  pom_dm_mult(JRa,set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler+3,rx,3,3,3,1);
  pom_dm_mult(JRb,set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler+3,ry,3,3,3,1);
  pom_dm_mult(JRg,set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.euler+3,rz,3,3,3,1);

  /** To be checked i mot sure */
  pom_dm_fill(Jto,6,6,
	  -1.0,0.0,0.0,0.0,0.0,0.0,
	  0.0,-1.0,0.0,0.0,0.0,0.0,
	  0.0,0.0,-1.0,0.0,0.0,0.0,
	  rx[0],ry[0],rz[0],-t[0],-t[1],-t[2] ,
	  rx[1],ry[1],rz[1],-t[4],-t[5],-t[6] ,
	  rx[2],ry[2],rz[2],-t[8],-t[9],-t[10]);

  pom_dmfsf_jmult_6x6(Jgt,tmp->euler.v, m1);
  pom_dmfsf_jmult_6x6(Jgo,set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.euler.v, m2);
  pom_dmsf_multtr_6x6(set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin.euler.v,Jto,m3);
  pom_dm_mult(Joo,m3,m3,6,6,6,6);
  pom_dm_mult(Jgo,m3,m3,6,6,6,6);
  pom_dm_trans_6x6(Jgt,Jgtr);
  pom_dm_mult(m3,Jgtr,m3,6,6,6,6);
  pom_dm_plustr_6x6(m3,m4);
  pom_dms_plus_6x6(m1,m2,m2);
  pom_dms_plus_6x6(m2,m4,m4);

    /** Come back to matrix representation */
  t3dConvertTo(T3D_U_MATRIX,&set->me[set->refMe].p[index1]->pos.pos.abs.mainToOrigin);
  t3dConvertTo(T3D_U_MATRIX,&set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin);
  t3dConvertTo(T3D_U_MATRIX,tmp);
  t3dConvertTo(T3D_U_MATRIX,displ);
  /** Last, the values composition */
  for(i=0;i<4;i++)
    for(j=0;j<4;j++)
      for (k=0, displ->matrix.matrix[i*4+j]=0.0;k<4;k++)
	displ->umatrix.matrix[i*4+j] += tmp->umatrix.matrix[i*4+k]*set->me[set->refMe].p[index2]->pos.pos.abs.mainToOrigin.umatrix.matrix[j+k*4];

  for (i=0;i<21;i++)
    displ->umatrix.v[i]=m4[i];

  t3dDelete(tmp);

  t3dConvertTo(T3D_EULER,displ);

  fprintf(stderr,"Dplacement entre %d et %d\n",index1,index2);

  return displ;

}
